@@ -30,10 +30,17 @@ from .wif   import WifEncoder
 
 @unique
 class Bip44Coins(IntEnum):
-    """ Enumerative for BIP44 coins. Only Bitcoin is present but it can be extended. """
+    """ Enumerative for BIP44 coins. Only some coins are present but it can be extended. """
 
-    BITCOIN         = 0,
-    BITCOIN_TESTNET = 1,
+    BITCOIN          = 0,
+    LITECOIN         = 2,
+    DOGECOIN         = 3,
+    ETHEREUM         = 60,
+    RIPPLE           = 144,
+    # Test nets. Special indexes are used here, they are converted to 1 internally
+    BITCOIN_TESTNET  = -1,
+    LITECOIN_TESTNET = -2,
+    DOGECOIN_TESTNET = -3,
 
 
 @unique
@@ -44,6 +51,23 @@ class Bip44Changes(IntEnum):
     CHAIN_INT = 1,
 
 
+@unique
+class Bip44PrivKeyTypes(IntEnum):
+    """ Enumerative for private key types. """
+
+    EXT_KEY = 0,
+    RAW_KEY = 1,
+
+
+@unique
+class Bip44PubKeyTypes(IntEnum):
+    """ Enumerative for public key types. """
+
+    EXT_KEY       = 0,
+    RAW_KEY       = 1,
+    RAW_COMPR_KEY = 2,
+
+
 class Bip44BaseConst:
     """ Class container for BIP44 base constants. """
 
@@ -60,6 +84,22 @@ class Bip44BaseConst:
     # Address depth
     ADDRESS_INDEX_DEPTH = 5
 
+    # Test ent coin index
+    TEST_NET_COIN_IDX   = 1
+
+    # Map test net coins to main net coins
+    TESTNET_TO_MAINNET_COINS = \
+        {
+            Bip44Coins.BITCOIN_TESTNET  : Bip44Coins.BITCOIN,
+            Bip44Coins.LITECOIN_TESTNET : Bip44Coins.LITECOIN,
+            Bip44Coins.DOGECOIN_TESTNET : Bip44Coins.DOGECOIN,
+        }
+
+
+class Bip44DepthError(Exception):
+    """ Expcetion in case of derivation from wrong depth. """
+    pass
+
 
 class Bip44Base(ABC):
     """ BIP44 base class.
@@ -67,42 +107,72 @@ class Bip44Base(ABC):
     The class is meant to be derived by classes implementing BIP44 or its extension.
     """
 
-    def __init__(self, bip32_obj):
-        """ Construct class from a Bip32 object.
+    def __init__(self, bip32_obj, coin_idx):
+        """ Construct class from a Bip32 object and coin type.
 
         Args:
             bip32_obj (Bip32 object) : Bip32 object
+            coin_idx (Bip44Coins)    : coin index, must be a Bip44Coins enum
         """
-        self.m_bip32 = bip32_obj
+        self.m_bip32    = bip32_obj
+        self.m_coin_idx = coin_idx
 
     @classmethod
-    def FromSeed(cls, seed_bytes, is_testnet = False):
+    def FromSeed(cls, seed_bytes, coin_idx):
         """ Create a Bip object (e.g. BIP44, BIP49, BIP84) from the specified seed (e.g. BIP39 seed).
         The test net flag is automatically set when the coin is derived. However, if you want to get the correct master
         or purpose keys, you have to specify here if it's a test net.
+        TypeError is raised if coin type is not a Bip44Coins enum.
 
         Args:
-            seed_bytes (bytes)          : seed bytes
-            is_testnet (bool, optional) : true if test net, false if main net (default value)
+            seed_bytes (bytes)    : seed bytes
+            coin_idx (Bip44Coins) : coin index, must be a Bip44Coins enum
 
         Returns (Bip object):
             Bip object
         """
-        return cls(Bip32.FromSeed(seed_bytes, is_testnet))
+
+        if not isinstance(coin_idx, Bip44Coins):
+            raise TypeError("Coin index is not an enumerative of Bip44Coins")
+
+        # Check if test net
+        if coin_idx in Bip44BaseConst.TESTNET_TO_MAINNET_COINS:
+            coin_idx = Bip44BaseConst.TESTNET_TO_MAINNET_COINS[coin_idx]
+            is_testnet = True
+        else:
+            is_testnet = False
+
+        return cls(Bip32.FromSeed(seed_bytes, is_testnet), coin_idx)
 
     @classmethod
-    def FromExtendedKey(cls, key_str):
+    def FromExtendedKey(cls, key_str, coin_idx):
         """ Create a Bip object (e.g. BIP44, BIP49, BIP84) from the specified extended key.
 
         Args:
-            key_str (str) : extended key string
+            key_str (str)         : extended key string
+            coin_idx (Bip44Coins) : coin index, must be a Bip44Coins enum
 
         Returns (Bip object):
             Bip object
         """
-        return cls(Bip32.FromExtendedKey(key_str, cls._GetPubNetVersions(), cls._GetPrivNetVersions()))
 
-    def PublicKey(self, extended = True):
+        if not isinstance(coin_idx, Bip44Coins):
+            raise TypeError("Coin index is not an enumerative of Bip44Coins")
+
+        # Check if test net
+        if coin_idx in Bip44BaseConst.TESTNET_TO_MAINNET_COINS:
+            coin_idx = Bip44BaseConst.TESTNET_TO_MAINNET_COINS[coin_idx]
+            is_testnet = True
+        else:
+            is_testnet = False
+
+        # Get net versions
+        main_net_ver = cls._GetMainNetVersions(coin_idx)
+        test_net_ver = cls._GetTestNetVersions(coin_idx)
+
+        return cls(Bip32.FromExtendedKey(key_str, is_testnet, main_net_ver, test_net_ver), coin_idx)
+
+    def PublicKey(self, key_type = Bip44PubKeyTypes.EXT_KEY):
         """ Return the public key.
 
         Args:
@@ -111,18 +181,40 @@ class Bip44Base(ABC):
         Returns (str):
             Public key
         """
-        return self.m_bip32.ExtendedPublicKey(self._GetPubNetVersions()) if extended else self.m_bip32.PublicKeyBytes()
 
-    def PrivateKey(self, extended = True):
+        if not isinstance(key_type, Bip44PubKeyTypes):
+            raise TypeError("Key type is not an enumerative of Bip44PubKeyTypes")
+
+        if key_type == Bip44PubKeyTypes.EXT_KEY:
+            # Get versions
+            main_pub_net_ver = self._GetMainNetVersions(self.m_coin_idx)["pub"]
+            test_pub_net_ver = self._GetTestNetVersions(self.m_coin_idx)["pub"]
+            # Get extendewd key
+            return self.m_bip32.ExtendedPublicKey(main_pub_net_ver, test_pub_net_ver)
+        else:
+            return self.m_bip32.PublicKeyBytes(key_type == Bip44PubKeyTypes.RAW_COMPR_KEY)
+
+    def PrivateKey(self, key_type = Bip44PrivKeyTypes.EXT_KEY):
         """ Return the private key.
 
         Args:
-            extended (bool) : if true, the extended key encoded in base58 will be returned, oitherwise the key bytes
+            key_type (Bip44PrivKeyTypes) : private key type
 
         Returns (str):
             Private key
         """
-        return self.m_bip32.ExtendedPrivateKey(self._GetPrivNetVersions()) if extended else self.m_bip32.PrivateKeyBytes()
+
+        if not isinstance(key_type, Bip44PrivKeyTypes):
+            raise TypeError("Key type is not an enumerative of Bip44PrivKeyTypes")
+
+        if key_type == Bip44PrivKeyTypes.EXT_KEY:
+            # Get versions
+            main_pub_net_ver = self._GetMainNetVersions(self.m_coin_idx)["priv"]
+            test_pub_net_ver = self._GetTestNetVersions(self.m_coin_idx)["priv"]
+            # Get extendewd key
+            return self.m_bip32.ExtendedPrivateKey(main_pub_net_ver, test_pub_net_ver)
+        else:
+            return self.m_bip32.PrivateKeyBytes()
 
     def Address(self):
         """ Return address related to the current public key.
@@ -130,7 +222,17 @@ class Bip44Base(ABC):
         Returns (str):
             Address string
         """
-        return self._GetAddressGenClass().ToAddress(self.m_bip32.PublicKeyBytes(), self.m_bip32.IsTestNet())
+        addr_fct = self._GetComputeAddressFct(self.m_coin_idx)
+        return addr_fct(self.m_bip32.PublicKey(), self.m_bip32.IsTestNet())
+
+    def WalletImportFormat(self):
+        """ Return the current private key encoded in WIF.
+
+        Returns (str):
+            Address string
+        """
+        wif_net_ver = self._GetWifNetVersions(self.m_coin_idx)
+        return WifEncoder.Encode(self.m_bip32.PrivateKeyBytes(), wif_net_ver["main"] if self.m_bip32.IsTestNet() else wif_net_ver["test"])
 
     def IsMasterLevel(self):
         """ Return if it's a master path.
@@ -180,14 +282,6 @@ class Bip44Base(ABC):
         """
         return self.m_bip32.Depth() == Bip44BaseConst.ADDRESS_INDEX_DEPTH
 
-    def WalletImportFormat(self):
-        """ Return the current private key encoded in WIF.
-
-        Returns (str):
-            Address string
-        """
-        return WifEncoder.Encode(self.m_bip32.PrivateKeyBytes())
-
     @abstractmethod
     def Purpose(self):
         """ Derive a child key from the purpose and return a new Bip object (e.g. BIP44, BIP49, BIP84).
@@ -199,13 +293,10 @@ class Bip44Base(ABC):
         pass
 
     @abstractmethod
-    def Coin(self, coin_idx):
-        """ Derive a child key from the specified coin type and return a new Bip object (e.g. BIP44, BIP49, BIP84).
+    def Coin(self):
+        """ Derive a child key from the coin type specified at construction and return a new Bip object (e.g. BIP44, BIP49, BIP84).
         It shall call the underlying _CoinGeneric method with the current object as parameter.
 
-        Args:
-            coin_idx (Bip44Coins) : coin index, must a Bip44Coins enum
-
         Returns (Bip object):
             Bip object
         """
@@ -262,33 +353,53 @@ class Bip44Base(ABC):
 
     @staticmethod
     @abstractmethod
-    def _GetPubNetVersions():
-        """ Get public net versions.
+    def _GetMainNetVersions(coin_idx):
+        """ Get main net versions.
 
-        Returns (tuple):
-            Private net versions (main net in index 0, test net in index 1)
+        Args:
+            coin_idx (Bip44Coins) : coin index, must be a Bip44Coins enum
+
+        Returns (dict):
+            Main net versions (public at key "pub", private at key "priv")
         """
         pass
 
     @staticmethod
     @abstractmethod
-    def _GetPrivNetVersions():
-        """ Get private net versions.
+    def _GetTestNetVersions(coin_idx):
+        """ Get test net versions.
 
-        Returns (tuple):
-            Private net versions (main net in index 0, test net in index 1)
+        Args:
+            coin_idx (Bip44Coins) : coin index, must be a Bip44Coins enum
+
+        Returns (dict):
+            Test net versions (public at key "pub", private at key "priv")
         """
         pass
 
     @staticmethod
     @abstractmethod
-    def _GetAddressGenClass():
-        """ Get address generator calls.
-        The class shall define the following static method:
-            ToAddress(pub_key_bytes, is_testnet = False)
+    def _GetComputeAddressFct(coin_idx):
+        """ Compute compute address function.
 
-        Returns (object):
-            Address generator class
+        Args:
+            coin_idx (Bip44Coins) : coin index, must be a Bip44Coins enum
+
+        Returns (function):
+            Compute address function
+        """
+        pass
+
+    @staticmethod
+    @abstractmethod
+    def _GetWifNetVersions(coin_idx):
+        """ Get WIF net versions.
+
+        Args:
+            coin_idx (Bip44Coins) : coin index, must be a Bip44Coins enum
+
+        Returns (dict):
+            WIF net versions (public at key "pub", private at key "priv")
         """
         pass
 
@@ -296,7 +407,7 @@ class Bip44Base(ABC):
     def _PurposeGeneric(cls, bip_obj):
         """ Derive a child key from the purpose and return a new Bip object (e.g. BIP44, BIP49, BIP84).
         It shall be called from a child class.
-        RuntimeError is raised is chain depth is not suitable for deriving keys.
+        Bip44DepthError is raised is chain depth is not suitable for deriving keys.
 
         Args:
             bip_obj (BIP object) : Bip object (e.g. BIP44, BIP49, BIP84)
@@ -306,41 +417,34 @@ class Bip44Base(ABC):
             Bip object
         """
         if not cls.IsMasterLevel(bip_obj):
-            raise RuntimeError("Current depth (%d) is not suitable for deriving purpose" % bip_obj.m_bip32.Depth())
+            raise Bip44DepthError("Current depth (%d) is not suitable for deriving purpose" % bip_obj.m_bip32.Depth())
 
-        return cls(bip_obj.m_bip32.ChildKey(cls._GetPurpose()))
+        return cls(bip_obj.m_bip32.ChildKey(cls._GetPurpose()), bip_obj.m_coin_idx)
 
     @classmethod
-    def _CoinGeneric(cls, bip_obj, coin_idx):
-        """ Derive a child key from the specified coin type and return a new Bip object (e.g. BIP44, BIP49, BIP84).
+    def _CoinGeneric(cls, bip_obj):
+        """ Derive a child key from the coin type specified at construction and return a new Bip object (e.g. BIP44, BIP49, BIP84).
         It shall be called from a child class.
-        TypeError is raised if coin type is not a Bip44Coins enum.
-        RuntimeError is raised is chain depth is not suitable for deriving keys.
+        Bip44DepthError is raised is chain depth is not suitable for deriving keys.
 
         Args:
             bip_obj (BIP object)  : Bip object (e.g. BIP44, BIP49, BIP84)
-            coin_idx (Bip44Coins) : coin index, must a Bip44Coins enum
 
         Returns (Bip object):
             Bip object
         """
-        if not isinstance(coin_idx, Bip44Coins):
-            raise TypeError("Coin index is not an enumerative of Bip44Coins")
-
         if not cls.IsPurposeLevel(bip_obj):
-            raise RuntimeError("Current depth (%d) is not suitable for deriving coin" % bip_obj.m_bip32.Depth())
+            raise Bip44DepthError("Current depth (%d) is not suitable for deriving coin" % bip_obj.m_bip32.Depth())
 
-        # Set test net depending on the coin
-        if coin_idx == Bip44Coins.BITCOIN_TESTNET:
-            bip_obj.m_bip32.SetTestNet(True)
+        coin_idx = Bip44BaseConst.TEST_NET_COIN_IDX if bip_obj.m_bip32.IsTestNet() else bip_obj.m_coin_idx
 
-        return cls(bip_obj.m_bip32.ChildKey(Bip32.HardenIndex(coin_idx)))
+        return cls(bip_obj.m_bip32.ChildKey(Bip32.HardenIndex(coin_idx)), bip_obj.m_coin_idx)
 
     @classmethod
     def _AccountGeneric(cls, bip_obj, acc_idx):
         """ Derive a child key from the specified account index and return a new Bip object (e.g. BIP44, BIP49, BIP84).
         It shall be called from a child class.
-        RuntimeError is raised is chain depth is not suitable for deriving keys.
+        Bip44DepthError is raised is chain depth is not suitable for deriving keys.
 
         Args:
             bip_obj (BIP object) : Bip object (e.g. BIP44, BIP49, BIP84)
@@ -350,16 +454,16 @@ class Bip44Base(ABC):
             Bip object
         """
         if not cls.IsCoinLevel(bip_obj):
-            raise RuntimeError("Current depth (%d) is not suitable for deriving account" % bip_obj.m_bip32.Depth())
+            raise Bip44DepthError("Current depth (%d) is not suitable for deriving account" % bip_obj.m_bip32.Depth())
 
-        return cls(bip_obj.m_bip32.ChildKey(Bip32.HardenIndex(acc_idx)))
+        return cls(bip_obj.m_bip32.ChildKey(Bip32.HardenIndex(acc_idx)), bip_obj.m_coin_idx)
 
     @classmethod
     def _ChangeGeneric(cls, bip_obj, chain_idx):
         """ Derive a child key from the specified chain type and return a new Bip object (e.g. BIP44, BIP49, BIP84).
         It shall be called from a child class.
         TypeError is raised if chain type is not a Bip44Changes enum.
-        RuntimeError is raised is chain depth is not suitable for deriving keys.
+        Bip44DepthError is raised is chain depth is not suitable for deriving keys.
 
         Args:
             bip_obj (BIP object)    : Bip object (e.g. BIP44, BIP49, BIP84)
@@ -372,15 +476,15 @@ class Bip44Base(ABC):
             raise TypeError("Chain index is not an enumerative of Bip44Changes")
 
         if not cls.IsAccountLevel(bip_obj):
-            raise RuntimeError("Current depth (%d) is not suitable for deriving chain" % bip_obj.m_bip32.Depth())
+            raise Bip44DepthError("Current depth (%d) is not suitable for deriving chain" % bip_obj.m_bip32.Depth())
 
-        return cls(bip_obj.m_bip32.ChildKey(chain_idx))
+        return cls(bip_obj.m_bip32.ChildKey(chain_idx), bip_obj.m_coin_idx)
 
     @classmethod
     def _AddressIndexGeneric(cls, bip_obj, addr_idx):
         """ Derive a child key from the specified address index and return a new Bip object (e.g. BIP44, BIP49, BIP84).
         It shall be called from a child class.
-        RuntimeError is raised is chain depth is not suitable for deriving keys.
+        Bip44DepthError is raised is chain depth is not suitable for deriving keys.
 
         Args:
             bip_obj (BIP object) : Bip object (e.g. BIP44, BIP49, BIP84)
@@ -390,6 +494,6 @@ class Bip44Base(ABC):
             Bip object
         """
         if not cls.IsChangeLevel(bip_obj):
-            raise RuntimeError("Current depth (%d) is not suitable for deriving address" % bip_obj.m_bip32.Depth())
+            raise Bip44DepthError("Current depth (%d) is not suitable for deriving address" % bip_obj.m_bip32.Depth())
 
-        return cls(bip_obj.m_bip32.ChildKey(addr_idx))
+        return cls(bip_obj.m_bip32.ChildKey(addr_idx), bip_obj.m_coin_idx)
